/**
 * Copyright (c) 2006-2009 Institut National de Recherche en Informatique et 
 *                         en Automatique. All rights reserved.
 * $COPYRIGHT$
 * 
 * Additional copyrights may follow
 * 
 * $HEADER$
 */

#include <stdio.h> 
#include <sys/types.h> 
#include <sys/socket.h> 
#include <netinet/in.h> 
#include <stdlib.h>
#include <string.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <netdb.h>
#include <signal.h>
#include <sys/select.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <netinet/tcp.h>
#include <netinet/ip.h>
#include <net/ethernet.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <pthread.h>

#include "cmdline.h"
#include "../proxy/util.h"
#include "frontale.h"
#include "debug.h"

#include "stubs/serviceH.h"
#include "stubs/service.nsmap"

char host_name[HOST_NAME_MAX];
char host_ip[MAX_IP];
int running;

static pthread_mutex_t nodes_mutex = PTHREAD_MUTEX_INITIALIZER;
struct node nodes[MAX_CONNECTED];
int nodes_count = 0;


struct soap broker_soap;
struct soap frontale_soap;
struct soap client_soap;

int masterfd = 0;
int clientfd = 0;
int broker_sockfd = 0;
char broker_addr[MAX_ADDR];


struct Namespace *namespaces;
static struct gengetopt_args_info ARGS;
static void setup_signals(void);
static int connectBroker(void);
static void* handleBroker(void *p_tsoap);
static void  * handleClientRequest(void *tmp);

static int doclose(struct soap *soap)
{
  printf("doclose master = %d socket = %d \n",soap -> master, soap -> socket );
  return SOAP_OK;
}

int main(int argc, char *argv[])
{

  struct hostent *host = NULL;
  pthread_t tid;
  struct soap *tsoap;
  pid_t pid, sid; 
  
  if( cmdline_parser (argc, argv, &ARGS) != 0)
    exit(1) ;
 if( ! ARGS.nodaemon_given ) {
    pid = fork();
  if (pid < 0) 
      exit(1);
  if (pid > 0) 
      exit(EXIT_SUCCESS); 
  umask(0);
  sid = setsid();
  if (sid < 0)
      exit(1);

  close(STDIN_FILENO);
  close(STDOUT_FILENO);
  close(STDERR_FILENO); 
  {
    FILE *f = fopen("/var/run/frontale.pid", "w");
    if( f != NULL ) {
	fprintf(f, "%d\n", getpid());
	fclose(f);
    } else {
	f = fopen("/tmp/frontale.pid", "w");
	if(f != NULL) {
	    fprintf(f, "%d\n", getpid());
	    fclose(f);
	}
    }
  } 
}
  setup_signals();
  if(gethostname(host_name, HOST_NAME_MAX) == -1)
    {
      debug("unable to resolve my hostname");
      return ERROR;
    }
   
  host = gethostbyname(host_name);
  if( host == NULL)
  {
      debug("gethostbyname failed");
      return ERROR;
  }
     
  strncpy(host_ip, inet_ntoa(*( struct in_addr*)( host->h_addr_list[0])), MAX_IP);
  strncpy(host_name, host->h_name, HOST_NAME_MAX);
  debug("Host_name = %s Host_ip = %s", host_name, host_ip);
  debug("1");   
  soap_init(&frontale_soap); 
  soap_set_namespaces(&frontale_soap, service_namespaces);
  debug("2"); 
  masterfd = soap_bind(&frontale_soap, NULL, 18000, 100); 
  if (masterfd < 0) 
  {
      soap_print_fault(&frontale_soap, stderr);
      return ERROR;
  }
  debug("4"); 
  debug("Before Connecting to broker");         
  if(connectBroker() < 0)
  {
     debug("connectBroker() failed");
      return ERROR;
  }
  debug("Connected to broker");
  
  frontale_soap.socket = broker_sockfd;
  frontale_soap.fclose = doclose;
  
  pthread_create(&tid, NULL, handleBroker, (void *)NULL);
  
  soap_init(&client_soap); 
  soap_set_namespaces(&client_soap, service_namespaces);
  clientfd = soap_bind(&client_soap, NULL, ARGS.client_port_arg, 100);
  running = 1;
  while(running)
  {
    debug("Wating for clients requests ..... ");
    client_soap.accept_timeout = 60;  // non blocking
    if(soap_accept(&client_soap) < 0) 
      continue;
    tsoap = soap_copy(&client_soap);
    pthread_create(&tid, NULL, handleClientRequest, (void*)tsoap);
    
  }
  soap_done(&client_soap);
  return SUCCESS;
   
  
  
  /*
   * soap_init(&broker_soap);
   * soap_set_namespaces(&broker_soap, service_namespaces);
   * sprintf(broker_addr,"http://%s:%d/",ARGS.broker_host_arg, ARGS.soap_port_arg);
   * soap_call_ns__update_broker(&broker_soap, broker_addr, "",&number,&out);
   * if(broker_soap.error)
   * {
   * soap_print_fault(&broker_soap, stderr);
   * printf("soap error\n");
   * }
   */
     
  return SUCCESS;
  
}
static void  * handleClientRequest(void *p_tsoap)
{
  struct soap *v_soap=(struct soap *)p_tsoap;
 
  pthread_detach(pthread_self());
  service_serve((struct soap*)v_soap);
  soap_destroy((struct soap*)v_soap);
  soap_end((struct soap*)v_soap);
  soap_done((struct soap*)v_soap);
  free(v_soap);
  return NULL;
 
}
void  * handleBroker(void *tmp)
{
  struct soap *tmpsoap;
  int run = 1;
  
  while(run)
  {
    tmpsoap = soap_copy(&frontale_soap);
    if (service_serve(tmpsoap) != SOAP_OK)
    {
      soap_print_fault(tmpsoap, stderr);
      run = 0;
    } 
    soap_destroy(tmpsoap);
    soap_end(tmpsoap);
    soap_done(tmpsoap);
  }
  pthread_exit(NULL);
}


void interrupt_catch(int signo)
{
  running = 0; 
}

void setup_signals(void)
{
  struct sigaction    sa;
  
  sa.sa_handler = SIG_IGN;                /* SIGPIPE to be ignored         */
  sigemptyset(&sa.sa_mask);               /* No signals to mask off        */
  sa.sa_flags = 0;                        /* No flags needed               */
 
  if (sigaction(SIGPIPE, &sa, NULL) == -1)
  {
    debug("sigaction(SIGPIPE)");
    exit(1);
  }
 
  sa.sa_handler = interrupt_catch;        /* SIGINT signal handler         */
  sigemptyset(&sa.sa_mask);               /* No signals to mask off        */
  sa.sa_flags = SA_RESTART;               /* Restart interrupted syscalls  */
 
  if (sigaction(SIGINT, &sa, NULL) == -1)
  {
    debug("sigaction(SIGINT)");
    exit(1);
  }
} 

int connectBroker(void)
{
  struct sockaddr_in serv_addr; 
  int nwrite;
  char buffer[MAX_BUFFER];
  struct hostent *host = NULL;
  
  
  sprintf(broker_addr,"http://%s:%d/",ARGS.broker_host_arg, ARGS.soap_port_arg); 
  if((broker_sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) 
  {
    debug("connectBroker: socket() failed:");
    return ERROR;
  }
  serv_addr.sin_family = AF_INET;
  host = gethostbyname( ARGS.broker_host_arg );
  if(host == NULL)
  {
   debug( "Unable to resolve broker hostname (%s): %s\n", ARGS.broker_host_arg, hstrerror(h_errno));
    return ERROR;
  }
  memcpy(&serv_addr.sin_addr, host->h_addr, host->h_length);
  serv_addr.sin_port = htons(ARGS.broker_port_arg);
  if (connect(broker_sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0) 
  {
   debug("Connect on broker at %s:%d failed: %s\n", inet_ntoa(serv_addr.sin_addr), ntohs(serv_addr.sin_port), strerror(errno));
    return ERROR;
  }
  sprintf(buffer,"%s %s", host_ip, host_name);
  if((nwrite = write(broker_sockfd, buffer, MAX_BUFFER)) <= 0)
  {
    debug("connectBroker: write() failed:");
    return ERROR;
  }
  return SUCCESS;
}
  
int ns__update_frontale(struct soap *soap1, int *number, struct ns__noResponse *out)
{
  printf("gsoap send me the number %d\n",*number);
  return SOAP_OK;
}
int ns__proxy_port(struct soap *p_soap, int *proxy_port, struct ns__noResponse  *out)
{
  return SOAP_OK;

}
int ns__getHostURI(struct soap *soap, process_name_t *proc_name, process_name_t *my_name, sockaddr_t *uri_A, TAG_t tag, ns__uriResponse *uri_B)
{
  struct soap brsoap;
  ns__uriResponse uri_rep;

  soap_init(&brsoap);
  soap_set_namespaces( &brsoap, service_namespaces);
  debug("frontale: soap_call_ns__getHostURI %s\n",broker_addr);

  soap_call_ns__getHostURI(&brsoap, broker_addr, "",proc_name, my_name, uri_A, tag, &uri_rep);
  if(brsoap.error)
    {
      soap_print_fault(&brsoap, stderr);
      return SOAP_ERR;;
    }
  debug( "ns__getHostURI {%s} {%s}:{%d}", uri_rep.host_name, uri_rep.host_ip, uri_rep.port);
  uri_B -> host_name = (char *)malloc(MAX_HOST_NAME * sizeof(char));
  memset(uri_B -> host_name, 0, MAX_HOST_NAME);
  uri_B -> host_ip = (char *)malloc(MAX_IP * sizeof(char));
  memset(uri_B ->host_ip, 0, MAX_IP);

  strncpy(uri_B -> host_ip, uri_rep.host_ip, MAX_IP);
  strncpy(uri_B -> host_name, uri_rep.host_name, MAX_HOST_NAME);
 
  if( uri_rep.response != NULL)
    {
      uri_B -> response = (char *)malloc(MAX_RESPONSE * sizeof(char));
      memset(uri_B -> response, 0, MAX_RESPONSE);
      strncpy(uri_B -> response, uri_rep.response, MAX_RESPONSE);
    }
  uri_B -> port = uri_rep.port;
  uri_B -> traversing_code = uri_rep.traversing_code;
  
  soap_end(&brsoap);
  return SOAP_OK;
}
int ns__putMyURI(struct soap *soap, process_name_t *proc_name, sockaddr_t *uri_B, TAG_t tag, struct ns__noResponse  *out)
{
  struct soap brsoap;
  soap_init(&brsoap);
  soap_set_namespaces( &brsoap, service_namespaces);
  pthread_mutex_lock(&nodes_mutex);
  if(proc_name == NULL)
      debug( "ns__putMyURI proc_name == NULL");
  else
      {
	  debug( "ns__putMyURI %d.%d.%d.%d %d", (int)proc_name->cellid,(int)proc_name->jobid,(int)proc_name->vpid, tag,  proc_name->taskid);
      }
  
  sscanf(uri_B -> response, "%s %s %d", nodes[nodes_count].node_id,nodes[nodes_count].hostname, &nodes[nodes_count].port );
  nodes_count++;
  pthread_mutex_unlock(&nodes_mutex);
  soap_call_ns__putMyURI(&brsoap, broker_addr, "", proc_name, uri_B, tag, out);
  if(brsoap.error)
    {
      soap_print_fault(&brsoap, stderr);
      return SOAP_ERR;
    }
  soap_end(&brsoap);
  return SOAP_OK;
}
int ns__update_broker(struct soap *soap, int *number, struct ns__noResponse  *out)
{
  return SOAP_OK;
}
int ns__update_connection_helper(struct soap *soap, int *number, struct ns__noResponse  *out)
{
  return SOAP_OK;
}

int ns__proxy_connection_demande(struct soap *soap, char *demande, struct ns__noResponse  *out)
{
  
  int i = 0 , found = 0;
  char proxy_ip[MAX_IP];
  char node_id[MAX_ID];
  char node_addr[MAX_ADDR];
  int proxy_port, uid;
  struct soap node_soap;
  char new_demande[MAX_RESPONSE];
  
  sscanf(demande, "%s %d %s %d", node_id, &uid, proxy_ip, &proxy_port);
  debug("frontale_ns__proxy_connection_demande: NODE_ID = %s", node_id);
  while((!found ) && (i < nodes_count))
  {
     if(strncmp(node_id, nodes[i].node_id, MAX_ID) == 0)
      found = 1;
     else
      i++;
  }  
  if(!found)
  {
    debug("NODE_ID %s not found", node_id);
    return SOAP_ERR;
  }
  soap_init(&node_soap);
  soap_set_namespaces( &node_soap, service_namespaces);
  
  sprintf(node_addr,"http://%s:%d/",nodes[i].hostname,nodes[i].port); 
  sprintf(new_demande,"%d %s %d", uid, proxy_ip, proxy_port);
  soap_call_ns__proxy_connection_demande(&node_soap, node_addr, "", new_demande, out);
  if(node_soap.error)
    {
      soap_print_fault(&node_soap, stderr);
      return SOAP_ERR;;
    }
  debug("Sended to connection helper %s", new_demande);
  soap_end(&node_soap);
  
  return SOAP_OK;
}
int ns__get_proxy_uid(struct soap *p_soap, int *uid)
{
  return SOAP_OK;
}
int ns__service_proxy_port(struct soap *p_soap, int *service_proxy_port, struct ns__noResponse  *out)
{
  return SOAP_OK;
}

int ns__forward_syn_data(struct soap *p_soap, char *data, struct ns__noResponse  *out)
{
    return SOAP_OK;
}

int ns__start_traversing(struct soap *p_soap, char *local_ip, int *port, struct ns__noResponse  *out)
{
  return SOAP_OK;
}
int  ns__delete_data(struct soap *p_soap,int *taskid)                                                              
{               
    struct soap brsoap; 
                                                                                                                                                       
    soap_init(&brsoap);                                                                                                                                
    soap_set_namespaces( &brsoap, service_namespaces);                                                                                                 
    debug("delete taskid = %d", *taskid);
    soap_call_ns__delete_data(&brsoap, broker_addr, "", taskid);                                                       
    if(brsoap.error)                                                                                                                                   
	{                                                                                                                                                
	    soap_print_fault(&brsoap, stderr);                                                                                                             
	    return SOAP_ERR;;                                                                                                                              
	}                                                                                                         
                                                                                                                   
    return SOAP_OK;                                                                                                
}
  /*******gsoap reverse
  {
    int masterfd;
    struct soap  *tmpsoap;
    
    soap_init(&frontale_soap); 
    soap_set_namespaces(&frontale_soap, service_namespaces);
    
    masterfd = soap_bind(&frontale_soap, NULL, 0, 100); 
    if (masterfd <= 0) 
      soap_print_fault(&frontale_soap, stderr); 
    else
    {  
      if(connectBroker() < 0)
      {
          debug("gethostbyname failed");
          return ERROR;
      }
      debug("Connected to broker");
      frontale_soap.socket = broker_sockfd;
      frontale_soap.fclose = doclose;
      
      debug("1");
      tmpsoap = soap_copy(&frontale_soap);
      
      if (service_serve(tmpsoap) != SOAP_OK)
      {
        debug("2");
        soap_print_fault(tmpsoap, stderr); // print error 
        
      } 
      else
      {
        debug("3");
       debug( "request served\n");
      }
      soap_destroy((struct soap*)tmpsoap);
      soap_end((struct soap*)tmpsoap);
      soap_done((struct soap*)tmpsoap);
      tmpsoap=soap_copy(&frontale_soap);
      
      if (service_serve(tmpsoap) != SOAP_OK)
      {
        debug("2");
        soap_print_fault(tmpsoap, stderr); // print error 
        
      } 
      else
      {
        debug("3");
       debug( "request served\n");
      }
      debug("4");
      soap_destroy((struct soap*)tmpsoap);
      soap_end((struct soap*)tmpsoap);
      soap_done((struct soap*)tmpsoap);
      
      soap_destroy(&frontale_soap);  // clean up class instances 
      soap_end(&frontale_soap);
  
  
  
    }   
  }*********/
  
  /*******end*********/
  
  /*******gsoap normal
  {
    int number = 6;
    
    struct ns__noResponse out;
    sleep(10);
    sprintf(broker_addr,"http://%s:%d/",ARGS.broker_host_arg, 5000);//ARGS.broker_port_arg);
    debug("gsoap broker = %s", broker_addr );
    
    soap_init(&broker_soap);
    soap_set_namespaces(&broker_soap, service_namespaces);
    soap_call_ns__update_broker(&broker_soap, broker_addr, "",&number,&out);
    if(broker_soap.error)
    {
       soap_print_fault(&broker_soap, stderr);
       printf("soap error\n");
    }
    number = 9;
     soap_call_ns__update_broker(&broker_soap, broker_addr, "",&number,&out);
    if(broker_soap.error)
    {
       soap_print_fault(&broker_soap, stderr);
       printf("soap error\n");
    }
    soap_end(&broker_soap);
  }*********/
  /*******end*********/
  

