# -*- shell-script -*-
#
# Copyright (c) 2008      The University of Tennessee and The University
#                         of Tennessee Research Foundation.  All rights
#                         reserved.
# Copyright (c) 2008      Institut National de Recherche en Informatique
#                         et Automatique. All rights reserved.
# $COPYRIGHT$
# 
# Additional copyrights may follow
# 
# $HEADER$
#

# check for taktuk
# OMPI_CHECK_TAKTUK(prefix, [action-if-found], [action-if-not-found])
# --------------------------------------------------------
AC_DEFUN([OMPI_CHECK_TAKTUK],[
    AC_ARG_WITH([taktuk],
        [AC_HELP_STRING([--with-taktuk(=DIR)],
             [Build TAKTUK support])])
    AC_ARG_WITH([taktuk-libdir],
       [AC_HELP_STRING([--with-taktuk-libdir=DIR],
             [Search for TAKTUK libraries in DIR])])

    # Defaults
    ompi_check_taktuk_dir_msg="compiler default"
    ompi_check_taktuk_libdir_msg="linker default"

    # Save directory names if supplied
    AS_IF([test ! -z "$with_taktuk" -a "$with_taktuk" != "yes"],
          [ompi_check_taktuk_dir="$with_taktuk"
           ompi_check_taktuk_dir_msg="$ompi_check_taktuk_dir (from --with-taktuk)"])
    AS_IF([test ! -z "$with_taktuk_libdir" -a "$with_taktuk_libdir" != "yes"],
          [ompi_check_taktuk_libdir="$with_taktuk_libdir"
           ompi_check_taktuk_libdir_msg="$ompi_check_taktuk_libdir (from --with-taktuk-libdir)"])

    # If no directories were specified, look for TAKTUK_LIBDIR,
    # TAKTUK_INCLUDEDIR, and/or TAKTUK_ENVDIR.
    AS_IF([test -z "$ompi_check_taktuk_dir" -a -z "$ompi_check_taktuk_libdir"],
          [AS_IF([test ! -z "$TAKTUK_ENVDIR" -a -z "$TAKTUK_LIBDIR" -a -f "$TAKTUK_ENVDIR/taktuk.conf"],
                 [TAKTUK_LIBDIR=`egrep ^TAKTUK_LIBDIR= $TAKTUK_ENVDIR/taktuk.conf | cut -d= -f2-`])
           AS_IF([test ! -z "$TAKTUK_ENVDIR" -a -z "$TAKTUK_INCLUDEDIR" -a -f "$TAKTUK_ENVDIR/taktuk.conf"],
                 [TAKTUK_INCLUDEDIR=`egrep ^TAKTUK_INCLUDEDIR= $TAKTUK_ENVDIR/taktuk.conf | cut -d= -f2-`])
           AS_IF([test ! -z "$TAKTUK_LIBDIR"],
                 [ompi_check_taktuk_libdir=$TAKTUK_LIBDIR
                  ompi_check_taktuk_libdir_msg="$TAKTUK_LIBDIR (from \$TAKTUK_LIBDIR)"])
           AS_IF([test ! -z "$TAKTUK_INCLUDEDIR"],
                 [ompi_check_taktuk_dir=`dirname $TAKTUK_INCLUDEDIR`
                  ompi_check_taktuk_dir_msg="$ompi_check_taktuk_dir (from \$TAKTUK_INCLUDEDIR)"])])

    AS_IF([test "$with_taktuk" = "no"],
          [ompi_check_taktuk_happy="no"],
          [ompi_check_taktuk_happy="yes"])

    ompi_check_taktuk_$1_save_CPPFLAGS="$CPPFLAGS"
    ompi_check_taktuk_$1_save_LDFLAGS="$LDFLAGS"
    ompi_check_taktuk_$1_save_LIBS="$LIBS"

    AS_IF([test "$ompi_check_taktuk_happy" = "yes"], 
          [AC_MSG_CHECKING([for TAKTUK dir])
           AC_MSG_RESULT([$ompi_check_taktuk_dir_msg])
           AC_MSG_CHECKING([for TAKTUK library dir])
           AC_MSG_RESULT([$ompi_check_taktuk_libdir_msg])
           OMPI_CHECK_PACKAGE([$1],
                              [taktuk.h],
                              [taktuk],
                              [taktuk_init_threads],
                              [],
                              [$ompi_check_taktuk_dir],
                              [$ompi_check_taktuk_libdir],
                              [ompi_check_taktuk_happy="yes"],
                              [ompi_check_taktuk_happy="no"])])

    CPPFLAGS="$ompi_check_taktuk_$1_save_CPPFLAGS"
    LDFLAGS="$ompi_check_taktuk_$1_save_LDFLAGS"
    LIBS="$ompi_check_taktuk_$1_save_LIBS"

    # Reset for the next time we're called
    ompi_check_taktuk_dir=
    ompi_check_taktuk_libdir=

    AS_IF([test "$ompi_check_taktuk_happy" = "yes"],
          [$2],
          [AS_IF([test ! -z "$with_taktuk" -a "$with_taktuk" != "no"],
                 [AC_MSG_WARN([TAKTUK support requested (via --with-taktuk) but not found.])
                  AC_MSG_ERROR([Aborting.])])
           $3])
])
