/**
 * Copyright (c) 2006-2009 Institut National de Recherche en Informatique et 
 *                         en Automatique. All rights reserved.
 *
 * $COPYRIGHT$
 * 
 * Additional copyrights may follow
 * 
 * $HEADER$
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <pthread.h>
#include <errno.h>
#include <signal.h>
#include <syslog.h>
#include <netdb.h>
#include <fcntl.h>
#include <sys/time.h>
#include <string.h>
#include <arpa/inet.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <signal.h>
#include <ctype.h>
#include <getopt.h>
#include <poll.h>
#include "stubs/serviceH.h"
#include "stubs/service.nsmap"

#include "proxy.h"
#include "util.h"
#include "debug.h"
#include "cmdline.h"

int listen_port = PROXY_PORT;
int running = 1;
int listenfd;
int global_uid = UID;
pthread_mutex_t uid_lock = PTHREAD_MUTEX_INITIALIZER;

static struct gengetopt_args_info ARGS;
struct Namespace *namespaces;

struct socket_id socket_ids[MAX_PROXY_CONNECTIONS];
/** poll_sockets is the array of all sockets
    sockets is the array of client sockets
    sockets_count represents the number of clients */
struct pollfd    poll_sockets[MAX_PROXY_CONNECTIONS+1];
struct pollfd   *sockets = poll_sockets + 1;
int sockets_count = 0;

struct proxy_connection *connections = NULL;
pthread_mutex_t connections_lock = PTHREAD_MUTEX_INITIALIZER;
struct soap v_soap;
static void* handleUID(void *arg);
static void finish (int sig);
static int  tcp_listen ();
static void handleNewConnection();
static void addClientSocket(int sockfd);
static void handleData(int i);
static void deleteClientSocket(int i);
static void getid(char *id, cellid_t *cellid, jobid_t *jobid, vpid_t *vpid, int *tag);
static void handleDeleteIds(int i, char *cmd_buff);
static void handleConnectionDemand(int i, char *cmd_buff);
static int  findSocket (cellid_t cellid, jobid_t jobid, vpid_t vpid, TAG_t tag);      
static void sendPort(int isClient, int sockfd, char *id_src, char *id_dst,int port, int pthread_uid);
static void *startproxy(void *arg);
static struct proxy_connection *addNewConnection(int pthread_uid);
static void  delConnection(struct proxy_connection *conn);
static void  handleNewClient(int index, char *buff);
static  void* handleRequest(void *p_tsoap);

int port_min = 0;
int port_max = 0;

int main (int argc, char *argv[])
{
  int c, ret;
  struct soap brsoap;
  char broker_addr[MAX_HOST_NAME];
  struct ns__noResponse out;
  pthread_t tid;
  pid_t pid, sid; 
  
  (void) signal (SIGINT, finish);

  if( cmdline_parser (argc, argv, &ARGS) != 0)
	  exit(1) ;
  if( ! ARGS.nodaemon_given )
    {
      pid = fork();   
      if (pid < 0)   
          exit(1); 
      if (pid > 0)
          exit(EXIT_SUCCESS);
      umask(0);  
      sid = setsid(); 
      if (sid < 0)
         exit(1); 
      close(STDIN_FILENO); 
      close(STDOUT_FILENO);
      close(STDERR_FILENO);
      {
        FILE *f = fopen("/var/run/proxy.pid", "w");
        if( f != NULL ) {
          fprintf(f, "%d\n", getpid());
          fclose(f);
	} else {
	    f = fopen("/tmp/proxy.pid", "w");
	    if(f != NULL) {
		fprintf(f, "%d\n", getpid());
		fclose(f);
	    }
        }
      }
    }
  
  listen_port = ARGS.proxy_port_arg;

  soap_init(&brsoap);
  soap_set_namespaces( &brsoap, service_namespaces);
  sprintf(broker_addr, "http://%s:%d/", ARGS.broker_host_arg, ARGS.broker_port_arg);
  fprintf (stderr, "proxy: ns__proxy_port %s\n",broker_addr);
  soap_call_ns__service_proxy_port(&brsoap, broker_addr, "", &ARGS.service_port_arg, &out);
  if(brsoap.error)
    {
      soap_print_fault(&brsoap, stderr);
      return SOAP_ERR;;
    }
  soap_call_ns__proxy_port(&brsoap, broker_addr, "", &ARGS.proxy_port_arg, &out);
  if(brsoap.error)
    {
      soap_print_fault(&brsoap, stderr);
      return SOAP_ERR;;
    }
  soap_end(&brsoap);
  
  soap_init(&v_soap);
  soap_set_namespaces(&v_soap, service_namespaces);

  if(soap_bind(&v_soap, NULL, ARGS.service_port_arg, 100) < 0)
    return EXIT_FAILURE;

  pthread_create(&tid, NULL, handleUID, (void *)NULL);
  
  if((listenfd = tcp_listen()) < 0)
    exit(1);

  fprintf (stderr, "Proxy started: listening on port [%d] \n", ARGS.proxy_port_arg);

  poll_sockets[0].fd = listenfd;
  poll_sockets[0].events = POLLIN;
  
  while (running)
  {
    if ((ret = poll(poll_sockets, sockets_count+1, -1)) < 0)
      {
	fprintf(stderr, "Poll was interrupted... (%s) Exiting\n",
		strerror(errno));
	break;
      }
    else 
      {
        if( poll_sockets[0].revents & POLLIN )
	  {
	    debug("Connection on proxy socket");
	    handleNewConnection();
	  }
        for(c = 0; c < sockets_count; c++)
	  {
	    if( sockets[c].revents & POLLIN )
	      {
		debug("Data on Client socket %d number %d", sockets[c].fd, c);
		handleData(c);
	      } 
	  }
      }
  }

  debug("Shutting down proxy");
  return 0;
}
static void* handleUID(void *arg)
{
  struct soap *v_tsoap; // clone of the SOAP context for the thread
  pthread_t tid;

  while(running)
    {
      debug("Wating for requests ..... ");
      v_soap.accept_timeout = 60;  // non blocking
      if(soap_accept(&v_soap) < 0)
        continue;
      v_tsoap=soap_copy(&v_soap);
      pthread_create(&tid, NULL, handleRequest, (void*)v_tsoap);
    }
  soap_done(&v_soap);
  pthread_exit(NULL);
}
static void* handleRequest(void *p_tsoap)
{
  struct soap *v_soap=(struct soap *)p_tsoap;

  pthread_detach(pthread_self());
  service_serve((struct soap*)v_soap);
  soap_destroy((struct soap*)v_soap);
  soap_end((struct soap*)v_soap);
  soap_done((struct soap*)v_soap);
  free(v_soap);
  pthread_exit(NULL);
}

void getid(char *id, cellid_t *cellid, jobid_t *jobid, vpid_t *vpid, int *tag)
{
  sscanf(id, "%d.%d.%d.%d", (int *)cellid, (int *)jobid, (int *)vpid, tag);
}
   
void handleProxyInit(int i, char *cmd_buff)
{
  char **args;
  int argc, j,k;

  getCmdArgs(cmd_buff, &args, &argc);
  k = 0;
  for(j = socket_ids[i].ids_count; j < (socket_ids[i].ids_count + argc); j++)
  {
    getid(args[k], &(socket_ids[i].ids[j].cellid), &socket_ids[i].ids[j].jobid, 
	  &socket_ids[i].ids[j].vpid, (int *)&socket_ids[i].ids[j].tag);
    k++;
  }
  socket_ids[i].ids_count = socket_ids[i].ids_count + argc; 
}
int ns__service_proxy_port(struct soap *p_soap, int *service_proxy_port, struct ns__noResponse  *out)
{
  return SOAP_OK;
}

void handleDeleteIds(int i, char *cmd_buff)
{
  char **args;
  int argc, j, k;
  cellid_t cellid;
  jobid_t jobid;
  vpid_t vpid;
  TAG_t tag;
  
  getCmdArgs(cmd_buff, &args, &argc);
  for(j = 0; j < argc; j++)
  {
    getid(args[j], &cellid, &jobid, &vpid, (int *)&tag);
    for(k = 0; k < socket_ids[i].ids_count; k++)
    {
       if((socket_ids[i].ids[k].cellid == cellid) && 
	  (socket_ids[i].ids[k].jobid == jobid) && 
	  (socket_ids[i].ids[k].vpid == vpid) && 
	  (socket_ids[i].ids[k].tag == tag))
	 {
	   int m;
	   for(m = k; m < socket_ids[i].ids_count - 1; m ++)
	     {
	       socket_ids[i].ids[k].cellid = socket_ids[i].ids[k+1].cellid;
	       socket_ids[i].ids[k].jobid = socket_ids[i].ids[k+1].jobid;
	       socket_ids[i].ids[k].vpid = socket_ids[i].ids[k+1].vpid;
	       socket_ids[i].ids[k].tag = socket_ids[i].ids[k+1].tag;
	     }
	   socket_ids[i].ids_count --;
	 }     
    }
  }
}

void handleData(int i)
{
  int nread;
  char cmd_buff[MAX_CMD_SIZE];
  int cmdCode;
  debug("handleData before read %d", sockets[i].fd );
  if((nread = readCmd(sockets[i].fd, cmd_buff)) <= 0)
    {
      debug("readCmd <=0");
      if (nread == 0)
	{
	  fprintf(stderr, "Client seems down, closing connection...\n");
	  deleteClientSocket(i);
	}
      else
	perror("readCmd failed");
      return;
    }
  debug("handlePeerComand : %s", cmd_buff);
  cmdCode = getCmdCode(cmd_buff);
  if (cmdCode == ERROR)  return;
  
  switch(cmdCode)  {
  case codeProxyInit :
    handleProxyInit(i, cmd_buff);
    break;
  case codeDeleteId :
    handleDeleteIds(i, cmd_buff);
    break;
  case codeConnectionDemand :
    handleConnectionDemand(i, cmd_buff);
    break;
  case codeUniqueId : 
    handleNewClient(i, cmd_buff);
    break;
  }   
}

void deleteClientSocket(int i)
{
    int k;
    sockets[i].fd = sockets[sockets_count-1].fd;
    sockets[i].events = sockets[sockets_count-1].events;
    sockets[i].revents = 0;
    socket_ids[i].ids_count = socket_ids[sockets_count-1].ids_count;
    for(k = 0; k < socket_ids[i].ids_count; k++)
        {
	  socket_ids[i].ids[k].cellid = socket_ids[sockets_count-1].ids[k].cellid;
	  socket_ids[i].ids[k].jobid = socket_ids[sockets_count-1].ids[k].jobid;
	  socket_ids[i].ids[k].vpid = socket_ids[sockets_count-1].ids[k].vpid;
        }
    sockets_count--;
}

void addClientSocket(int sockfd)
{
  sockets[sockets_count].fd = sockfd;
  sockets[sockets_count].events = POLLIN;
  sockets[sockets_count].revents = 0;
  socket_ids[sockets_count].ids_count = 0;
  sockets_count++;
}

void handleNewConnection()
{
  int sockfd;
  struct sockaddr_in addr;  
  socklen_t len;     

  /* Set the size of the in-out parameter */
  len = sizeof(addr);
     
  /* Wait for a client to connect */
  if ((sockfd = accept(listenfd, (struct sockaddr *) &addr, &len)) < 0)
    {
      perror("accept failed");
      return;
    }
  debug("New client on socket %d accepted %s", sockfd, inet_ntoa(addr.sin_addr));
  addClientSocket(sockfd);
}

int tcp_listen ()
{
  int sock;                        /* socket to create */
  struct sockaddr_in brockAddr;   /* Local address */

  /* Create socket for incoming connections */
  if ((sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
  {
    perror("socket failed");
    return ERROR;
  }
  /* Construct local address structure */
  memset(&brockAddr, 0, sizeof(brockAddr));   /* Zero out structure */
  brockAddr.sin_family = AF_INET;             /* Internet address family */
  brockAddr.sin_addr.s_addr = INADDR_ANY; 
  brockAddr.sin_port = htons(listen_port);    /* Local port */

  /* Bind to the local address */
  if (bind(sock, (struct sockaddr *) &brockAddr, sizeof(brockAddr)) < 0)
  {
    perror("bind failed");
    return ERROR;
  }
  
  /* Mark the socket so it will listen for incoming connections */
  if (listen(sock, MAXPENDING) < 0)
  {
    perror("unable to put the socket in listen mode");
    return ERROR;
  }
  return sock;
}


void finish (int sig)
{
  pthread_exit(NULL);
}

void handleConnectionDemand(int i, char *cmd_buff)
{
    int  argc, targetfd, pthread_uid, ret;
    char **args;
  
    cellid_t cellid;
    jobid_t jobid;
    vpid_t vpid;
    TAG_t tag;
    pthread_t tid;
    struct proxy_connection *newcon;
    
    getCmdArgs(cmd_buff, &args, &argc); 
    if(argc < 1)
	     return;
    
    pthread_uid = global_uid;
    newcon = addNewConnection(pthread_uid);
    global_uid++;

    debug("Creating thread with uid %d", newcon->uid);
    if((ret = pthread_create(&tid, NULL, startproxy, (void *)newcon)) != 0)
      {
	perror("unable to create a posix thread");
	exit(1);
      }
  
    debug("Getting id from %s", args[1]);
    getid(args[1], &cellid, &jobid, &vpid, (int *)&tag);
    if((targetfd = findSocket(cellid, jobid, vpid, tag)) < 0)
      {
	fprintf(stderr, "The process %d.%d.%d.%d is not connected. Cannot link processes (FATAL)\n",
		(int )cellid, (int )jobid, (int )vpid, (int )tag);
	exit(1);
      }
    sendPort(1, sockets[i].fd, args[0], args[1], listen_port, pthread_uid);
    sendPort(0, targetfd, args[0], args[1], listen_port, pthread_uid);
}

int findSocket (cellid_t cellid, jobid_t jobid, vpid_t vpid, TAG_t tag)
{
  int i = 0, j = 0;
  
  for(i = 0; i < sockets_count; i++)
  {
    for(j = 0; j < socket_ids[i].ids_count; j++)
      {
	if((socket_ids[i].ids[j].cellid == cellid) &&
	   (socket_ids[i].ids[j].jobid == jobid) && 
	   (socket_ids[i].ids[j].vpid == vpid) && 
	   (socket_ids[i].ids[j].tag == tag))
	  return sockets[i].fd;
      }
  }
  return ERROR;
}

void sendPort(int isClient, int sockfd, char *id_src, char *id_dst,int port, int pthread_uid)
{
  char buff[MAX_CMD_SIZE];
  int nwrite = 0;  
  if(isClient)
      sprintf(buff,"%d %s %s %d %d", codeConnectionDemandReponse, id_src, id_dst, port, pthread_uid);
  else
      sprintf(buff,"%d %s %s %d %d", codeConnectionDemand, id_src, id_dst, port, pthread_uid);
  debug("Sending port %s", buff);
  nwrite = writeCmd(sockfd, buff);
  debug("Sent port %s (%d bytes)", buff, nwrite);
}

void *startproxy(void *arg)
{
    int uid, client, server;
    struct proxy_connection *myconnection = (struct proxy_connection*)arg;
    fd_set sockSet;
    struct sockaddr_in client_addr;
    struct sockaddr_in server_addr;
    char client_straddr[64], server_straddr[64];
    socklen_t slen;
    char * buf;

    pthread_mutex_lock(&(myconnection->mutex));
    while(( myconnection->client == -1 )||(myconnection->server == -1))
      pthread_cond_wait(&(myconnection -> cond), &(myconnection -> mutex));
    pthread_mutex_unlock(&(myconnection -> mutex));    
    
    client = myconnection->client;
    server = myconnection->server;
    uid = myconnection->uid;

    slen = sizeof(struct sockaddr_in);
    getpeername(client, (struct sockaddr*)&client_addr, &slen);
    snprintf(client_straddr, 64, "%s:%d", inet_ntoa(client_addr.sin_addr), ntohs(client_addr.sin_port));
    getpeername(server, (struct sockaddr*)&server_addr, &slen);
    snprintf(server_straddr, 64, "%s:%d", inet_ntoa(server_addr.sin_addr), ntohs(server_addr.sin_port));

    debug("connection of uid %d links %s to %s. Beginning proxy",
	  uid, client_straddr, server_straddr);

    delConnection(myconnection);
    
    buf = (char*)malloc(TRANSFER_BUFFER_SIZE);

    while (1)
    {
	int ret;
	/* Zero socket descriptor vector and set for brocker sockets */
	/* This must be reset every time select() is called */
	FD_ZERO(&sockSet);
	FD_SET(server, &sockSet);
	FD_SET(client, &sockSet);
	
	if ((ret = select(FD_SETSIZE, &sockSet, NULL, NULL, NULL)) < 0){
	    debug("Closing connections");
	    close(client);
	    close(server);
	    break;
	}

	if(FD_ISSET(client, &sockSet)){
	    int nbytes;
	    nbytes = read(client, buf, TRANSFER_BUFFER_SIZE);
	    /* debug("c-%d bytes->s on connection %d from %s to %s: '%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c'",
		  nbytes, uid, client_straddr, server_straddr,
		  isalpha(buf[0])?buf[0]:'.',
		  isalpha(buf[1])?buf[1]:'.',
		  isalpha(buf[2])?buf[2]:'.',
		  isalpha(buf[3])?buf[3]:'.',
		  isalpha(buf[4])?buf[4]:'.',
		  isalpha(buf[5])?buf[5]:'.',
		  isalpha(buf[6])?buf[6]:'.',
		  isalpha(buf[7])?buf[7]:'.',
		  isalpha(buf[8])?buf[8]:'.',
		  isalpha(buf[9])?buf[9]:'.',
		  isalpha(buf[10])?buf[10]:'.',
		  isalpha(buf[11])?buf[11]:'.',
		  isalpha(buf[12])?buf[12]:'.',
		  isalpha(buf[13])?buf[13]:'.',
		  isalpha(buf[14])?buf[14]:'.',
		  isalpha(buf[15])?buf[15]:'.');*/
		  
	    if (nbytes <= 0){
		perror("proxy read failed");
		close(client);
		close(server);
		break;
	    }
	    proxyWrite(server, buf, nbytes);
	}
	if(FD_ISSET(server, &sockSet)){
	    int nbytes;
	    nbytes = read(server, buf, TRANSFER_BUFFER_SIZE);
	    /*debug("c<-%d bytes-s on connection %d from %s to %s: '%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c'",
		  nbytes, uid, server_straddr, client_straddr,
		  isalpha(buf[0])?buf[0]:'.',
		  isalpha(buf[1])?buf[1]:'.',
		  isalpha(buf[2])?buf[2]:'.',
		  isalpha(buf[3])?buf[3]:'.',
		  isalpha(buf[4])?buf[4]:'.',
		  isalpha(buf[5])?buf[5]:'.',
		  isalpha(buf[6])?buf[6]:'.',
		  isalpha(buf[7])?buf[7]:'.',
		  isalpha(buf[8])?buf[8]:'.',
		  isalpha(buf[9])?buf[9]:'.',
		  isalpha(buf[10])?buf[10]:'.',
		  isalpha(buf[11])?buf[11]:'.',
		  isalpha(buf[12])?buf[12]:'.',
		  isalpha(buf[13])?buf[13]:'.',
		  isalpha(buf[14])?buf[14]:'.',
		  isalpha(buf[15])?buf[15]:'.');
	    */  
	    if (nbytes <= 0){
		close(client);
		close(server);
		break;
	    }
	    proxyWrite(client, buf, nbytes);
	}    
    }

    free(buf);
    pthread_exit(NULL);
}

void freeConnections()
{
    struct proxy_connection *tmp;

    pthread_mutex_lock(&connections_lock);
    while(connections != NULL)
    {
	tmp = connections->next;
	pthread_mutex_destroy(&connections->mutex);
	pthread_cond_destroy(&connections->cond);
	free(connections);
	connections = tmp;
    }
    pthread_mutex_lock(&connections_lock);
}

struct proxy_connection *addNewConnection(int pthread_uid)
{
  struct proxy_connection *tmp;

  tmp = (struct proxy_connection *)malloc(sizeof(struct proxy_connection));
  tmp->client = -1;
  tmp->server = -1;
  tmp->uid = pthread_uid;
  pthread_mutex_init(&tmp->mutex, NULL);
  pthread_cond_init(&tmp->cond, NULL);
  
  pthread_mutex_lock(&connections_lock);
  tmp->next = connections;
  connections = tmp;
  pthread_mutex_unlock(&connections_lock);
  
  return tmp;
}

void delConnection(struct proxy_connection *c)
{
  struct proxy_connection *tmp;

  pthread_mutex_lock(&connections_lock);
  if( connections == c )
    connections = connections->next;
  else
    {
      tmp = connections;
      while( (NULL != tmp->next) && (tmp->next != c) )
	tmp = tmp->next;
      if( NULL == tmp->next )
	fprintf(stderr, "delConnection: connection %d does not exists...\n",
		c->uid);
      else
	tmp->next = c->next;  
    }
  pthread_mutex_unlock(&connections_lock);
  c->next = NULL;
  pthread_cond_destroy(&c->cond);
  pthread_mutex_destroy(&c->mutex);
  free(c);
}

void handleNewClient(int index, char *cmd_buff)
{
    int  argc, uid;
    char **args;
    struct sockaddr_in addr;
    socklen_t slen;
    struct proxy_connection *newconnection;
    
    getCmdArgs(cmd_buff, &args, &argc);
    if(argc < 1) 
	return;
    uid = atoi(args[0]);

    slen = sizeof(struct sockaddr_in);
    getpeername(sockets[index].fd, (struct sockaddr*)&addr, &slen);
    
    pthread_mutex_lock(&connections_lock);
    newconnection = connections;
    while( (newconnection != NULL) && (newconnection->uid != uid))
      {
	debug("handleNewClient: uid = %d", newconnection->uid); 
	newconnection = newconnection -> next;
      }
    pthread_mutex_unlock(&connections_lock);

    if(newconnection == NULL){
	debug("handleNewClient: could not find uid %d", uid);
	return;
    }
    
    pthread_mutex_lock(&(newconnection->mutex));    
    if(newconnection -> client != -1)
      {
	newconnection->server = sockets[index].fd;
	debug("handleNewClient: %s:%d is second peer of %d, signaling proxy", 
	      inet_ntoa(addr.sin_addr), ntohs(addr.sin_port), uid);
	pthread_cond_broadcast(&(newconnection->cond));
      }
    else
      {
	newconnection -> client = sockets[index].fd;
	debug("handleNewClient: %s:%d is first peer of %d", 
	      inet_ntoa(addr.sin_addr), ntohs(addr.sin_port), uid);
      }
    pthread_mutex_unlock(&(newconnection -> mutex));
    deleteClientSocket(index);
}



int ns__getHostURI(struct soap *p_soap, process_name_t *proc_name, process_name_t *my_name, sockaddr_t *uri_A, TAG_t tag, ns__uriResponse *uri_B)
{
  return SOAP_OK;
}
int ns__putMyURI(struct soap *p_soap, process_name_t *proc_name, sockaddr_t *uri_B, TAG_t tag, struct ns__noResponse  *out)
{
  return SOAP_OK;
}
int ns__update_broker(struct soap *p_soap, int *number, struct ns__noResponse  *out)
{
  return SOAP_OK;
}
int ns__update_frontale(struct soap *p_soap, int *number, struct ns__noResponse  *out)
{
  return SOAP_OK;
}
int ns__update_connection_helper(struct soap *p_soap, int *number, struct ns__noResponse  *out)
{
  return SOAP_OK;
}
int ns__proxy_connection_demande(struct soap *p_soap, char *demande, struct ns__noResponse  *out)
{
  return SOAP_OK;
}
int ns__proxy_port(struct soap *p_soap, int *proxy_port, struct ns__noResponse  *out)
{
  return SOAP_OK;
}
int ns__get_proxy_uid(struct soap *p_soap, int *uid)
{
  int local_uid = 0;
  int ret = 0;
  struct proxy_connection *newcon;
  pthread_t tid;
  
  pthread_mutex_lock(&uid_lock);
  local_uid = global_uid;
  *uid = local_uid;
  
  global_uid++;
  newcon = addNewConnection(local_uid);
  debug("Creating thread with uid %d %d", newcon->uid, *uid);
  if((ret = pthread_create(&tid, NULL, startproxy, (void *)newcon)) != 0)
  {
    perror("unable to create a posix thread");
    return SOAP_ERR;
  }
  pthread_mutex_unlock(&uid_lock);
  return SOAP_OK;
}

int ns__forward_syn_data(struct soap *p_soap, char *data, struct ns__noResponse  *out)
{
    return SOAP_OK;
}

int ns__start_traversing(struct soap *p_soap, char *local_ip, int *port, struct ns__noResponse  *out)
{
  return SOAP_OK;
}


int  ns__delete_data(struct soap *p_soap,int *taskid)
{

    return SOAP_OK;
}   
